import pandas as pd
import logging


def flatten_bronze_table(df):

    df = df.reset_index()

    s = (df.applymap(type) == dict).all()
    dict_columns = s[s].index.tolist()

    while len(dict_columns) > 0:
        new_columns = []

        for col in dict_columns:
            # explode dictionaries horizontally, adding new columns
            horiz_exploded = pd.json_normalize(df[col]).add_prefix(f'{col}_')
            horiz_exploded.index = df.index
            df = pd.concat([df, horiz_exploded], axis=1).drop(columns=[col])
            new_columns.extend(horiz_exploded.columns)  # inplace

        # check if there are still dict fields to flatten
        s = (df[new_columns].applymap(type) == dict).all()
        dict_columns = s[s].index.tolist()

    logging.info("flatten_bronze_table: Returning flattened table.")
    return df.set_index("index")



from snowflake.connector.pandas_tools import write_pandas
import pandas as pd
import logging


def put_tables_into_snowflake(snowflake_connector, your_tables):
    """Takes in your SF Connector, and a dictionary of tables (key = table name) to deposit into Snowflake."""
    cs = snowflake_connector.cursor()

    for table_name in your_tables.keys():
        try:
            sql_command = pd.io.sql.get_schema(your_tables[table_name],
                                               table_name)
            insertion_index = sql_command.find("TABLE")
            sql_command = sql_command[:
                                      insertion_index] + "OR REPLACE " + sql_command[
                                          insertion_index:]
            cs.execute(sql_command)
            success, nchunks, nrows, _ = write_pandas(snowflake_connector,
                                                      your_tables[table_name],
                                                      table_name)
        finally:
            cs.execute("SELECT * FROM " + table_name)
            logging.info(
                "Finished writing tables to Snowflake, confirmed Select * works on table."
            )
    cs.close()
    snowflake_connector.close()



import pandas as pd
import labelsnow
import logging
from labelsnow.flatten_bronze_table import flatten_bronze_table
from labelsnow.add_json_answers_to_dictionary import add_json_answers_to_dictionary

def silver_table(df):
    flattened_bronze = flatten_bronze_table(df)

    # search for columns to explode/flatten
    s = (flattened_bronze.applymap(type) == list).all()
    list_columns = s[s].index.tolist(
    )  #generally yields ['Label_objects', 'Label_classifications', 'Label_relationships']

    video = False  #this will be used for future video frame handling
    if "Label_frameNumber" in flattened_bronze.columns:
        video = True

    new_json = []
    for index, row in flattened_bronze.iterrows():
        my_dictionary = {}

        # classifications
        try:  # this won't work if there are no classifications
            for index, classification_json in enumerate(
                    row["Label_classifications"]):
                title = classification_json["title"]
                if "answer" in classification_json:
                    if classification_json[
                            "answer"] is not None:  # if answer is null, that means it exists in secondary "answers" column
                        if "title" in classification_json["answer"]:
                            answer = classification_json["answer"]["title"]
                        else: #this is an edge case where the answer isn't nested in another layer; e.g. for our textfields
                            answer = classification_json["answer"]
                    else:
                        answer = classification_json[
                            "answers"] #This is for checklists or dropdowns
                else:
                    print(
                        "This line may be unnecessary"
                    )  #answer = row["Label.classifications.answer.title"][index]
                my_dictionary = add_json_answers_to_dictionary(
                    title, answer, my_dictionary)
        except Exception as e:
            print(e)

        # object counting
        try:  # this field won't work if the Label does not have objects in it
            for object in row.get("Label_objects", []):
                object_name = object["title"] + "_count"
                if object_name not in my_dictionary:
                    my_dictionary[object_name] = 1  # initialize with 1
                else:
                    my_dictionary[object_name] += 1  # add 1 to counter
        except Exception as e:
            print("No objects found.")

        #object classifications (if applicable). This is messy code which duplicates some above behavior, can be improved.
        try:  # this won't work if there are no classifications
            for object in row.get("Label_objects", []):
                if "classifications" in object:
                    for index, classification_json in enumerate(
                            object["classifications"]):
                        title = classification_json["title"]
                        if "answer" in classification_json:
                            if classification_json[
                                "answer"] is not None:  # if answer is null, that means it exists in secondary "answers" column
                                if "title" in classification_json["answer"]:
                                    answer = classification_json["answer"]["title"]
                                else:  # this is an edge case where the answer isn't nested in another layer; e.g. for our textfields
                                    answer = classification_json["answer"]
                            else:
                                answer = classification_json[
                                    "answers"]  # This is for checklists or dropdowns
                        else:
                            print(
                                "This line may be unnecessary"
                            )  # answer = row["Label.classifications.answer.title"][index]
                        my_dictionary = add_json_answers_to_dictionary(
                            title, answer, my_dictionary)
        except Exception as e:
            print(e)

        my_dictionary["DataRow ID"] = row["DataRow ID"]  # close it out
        if video:
            my_dictionary["Label_frameNumber"] = row[
                "Label_frameNumber"]  # need to store the unique framenumber identifier for video
        new_json.append(my_dictionary)

    parsed_classifications = pd.DataFrame(new_json)

    if video:
        # need to inner-join with frameNumber to avoid creating N-squared datarows, since each frame has same DataRowID
        joined_df = pd.merge(parsed_classifications,
                             flattened_bronze,
                             how='inner',
                             on=["DataRow ID", "Label_frameNumber"])
    else:
        joined_df = pd.merge(parsed_classifications,
                             flattened_bronze,
                             how='inner',
                             on="DataRow ID")
        # joined_df = parsed_classifications.join(flattened_bronze, ["DataRow ID"],
        #                                         "inner")

    return joined_df



import requests
import json
import pandas as pd
from .constants import LABELBOX_DEFAULT_TYPE_DICTIONARY


def get_videoframe_annotations(bronze_video_labels, api_key):
    # This method takes in the bronze table from get_annotations and produces
    # an array of bronze dataframes containing frame labels for each project
    # bronze_video_labels = bronze_video_labels.withColumnRenamed(
    #     "DataRow ID", "DataRowID")

    # We manually build a string of frame responses to leverage our existing jsonToDataFrame code, which takes in JSON
    headers = {'Authorization': f"Bearer {api_key}"}
    master_array_of_json_arrays = []
    for index, row in bronze_video_labels.iterrows():
        response = requests.get(row.Label["frames"],
                                headers=headers,
                                stream=False)
        data = []
        for line in response.iter_lines():
            data.append({
                "DataRow ID": row["DataRow ID"],
                "Label": json.loads(line.decode('utf-8'))
            })
        massive_string_of_responses = json.dumps(data)
        master_array_of_json_arrays.append(massive_string_of_responses)

    array_of_bronze_dataframes = []
    for frameset in master_array_of_json_arrays:
        data = json.loads(frameset)  #parse the JSON into a dict
        # df = pd.json_normalize(data) #had to use this b/c the data is a list of json objects
        # df = df.astype(LABELBOX_DEFAULT_TYPE_DICTIONARY)
        df = pd.DataFrame.from_dict(data).astype(
            {'DataRow ID': 'string'})  #create pandas DF with proper col type
        array_of_bronze_dataframes.append(df)  #create array of bronze tables

    return array_of_bronze_dataframes



LABELBOX_DEFAULT_TYPE_DICTIONARY = {
    'ID': 'string',
    'DataRow ID': 'string',
    'Labeled Data': 'string',
    'Created By': 'string',
    'Project Name': 'string',
    'Seconds to Label': 'float64',
    'External ID': 'string',
    'Agreement': 'int64',
    'Benchmark Agreement': 'int64',
    'Benchmark ID': 'string',
    'Dataset Name': 'string',
    'Reviews': 'object',
    'View Label': 'string',
    'Has Open Issues': 'int64',
    'Skipped': 'bool'
}


import logging


def create_dataset(labelbox_client, snowflake_pandas_dataframe, dataset_name):
    """Takes in a dataframe with the following column names: external_id, row_data
    # external_id is the asset name ex: "photo.jpg"
    # row_data is the signed URL to the asset
    returns: Labelbox client dataset object
    """
    dataSet_new = labelbox_client.create_dataset(name=dataset_name)
    snowflake_pandas_dataframe.columns = snowflake_pandas_dataframe.columns.str.lower(
    )
    data_row_urls = [{
        "external_id": row['external_id'],
        "row_data": row['row_data']
    } for index, row in snowflake_pandas_dataframe.iterrows()]
    upload_task = dataSet_new.create_data_rows(data_row_urls)
    upload_task.wait_till_done()
    logging.info("{}: Dataset creation. Dataset ID: {}".format(
        upload_task.status, dataSet_new.uid))

    return dataSet_new



from .add_json_answers_to_dictionary import add_json_answers_to_dictionary
from .constants import LABELBOX_DEFAULT_TYPE_DICTIONARY
from .create_dataset import create_dataset
from .flatten_bronze_table import flatten_bronze_table
from .get_annotations import get_annotations
from .get_snowflake_datarows import get_snowflake_datarows
from .is_json import is_json
from .silver_table import silver_table
from .get_videoframe_annotations import get_videoframe_annotations
from .put_tables_into_snowflake import put_tables_into_snowflake


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="labelsnow",
    version="0.1.2",
    author="Labelbox",
    author_email="ecosystem+snowflake@labelbox.com",
    description="Labelbox Connector for Snowflake",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://labelbox.com",
    packages=setuptools.find_packages(),
    install_requires=["labelbox", "pandas", "snowflake-connector-python"],
    keywords=["labelbox", "labelsnow"],
)



import logging


def get_snowflake_datarows(snowflake_cursor,
                           stage_name,
                           signed_url_expiration=604800):
    sql_string = (
        "select relative_path as external_id, "
        "get_presigned_url(@{s_name}, relative_path, {s_expiration}) as row_data "
        "from directory(@{s_name})".format(s_name=stage_name,
                                           s_expiration=signed_url_expiration))
    snowflake_cursor.execute(sql_string)
    logging.info("Executing query on Snowflake: " + sql_string)
    return snowflake_cursor.fetch_pandas_all()



import json
import urllib
import pandas as pd
import logging
from labelsnow.constants import LABELBOX_DEFAULT_TYPE_DICTIONARY


def get_annotations(labelbox_client, project_id):
    """Request annotations for a specific project_id and produce a Snowflake-ready Pandas Dataframe"""
    project = labelbox_client.get_project(project_id)
    with urllib.request.urlopen(project.export_labels()) as url:
        api_response_string = url.read().decode()  # this is a string of JSONs

    data = json.loads(api_response_string)
    df = pd.DataFrame.from_dict(data).astype(LABELBOX_DEFAULT_TYPE_DICTIONARY)

    #For some reason dtype dict does not convert timestamp reliably, so I must include these manual conversions
    df["Created At"] = pd.to_datetime(df["Created At"])
    df["Updated At"] = pd.to_datetime(df["Updated At"])

    logging.info("Returning annotations DataFrame from Labelbox")
    return df
